# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import logging
from pathlib import Path
from typing import (
    Any,
    Dict,
    Generator,
    TypeVar,
    Union,
)

from megatron.energon.epathlib import EPath
from megatron.energon.flavors.jsonl.ijsonl import IJsonlFile, IJsonlIndexWriter

logger = logging.getLogger(__name__)

T = TypeVar("T", covariant=True)


class JsonlPreparator:
    @staticmethod
    def iter_dataset_content(
        path: Union[str, EPath],
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Yield example dataset content for a few samples.

        Args:
            path: Path to the tar file.
        """
        with EPath(path).open("rb") as f:
            with IJsonlFile(f) as index_reader:
                for entry in index_reader:
                    yield {"json": entry}

    @classmethod
    def prepare_dataset(
        cls,
        path: Union[Path, EPath],
    ) -> int:
        """
        Preprocess the jsonl file. Preprocessing is done in parallel.
        Counts the number of samples.

        Args:
            path: Path to the jsonl file

        Returns:
            Count of samples in the jsonl file.
        """
        count = 0
        # Processing is lagging behind. The offsets include empty lines. The whole file must be covered!
        last_offset = 0
        with IJsonlIndexWriter(EPath(path)) as iw:
            with EPath(path).open("rb") as f:
                while True:
                    line = f.readline()
                    if not line:
                        break
                    line = line.strip()
                    if not line:
                        if last_offset:
                            last_offset = f.tell()
                        continue
                    assert line.startswith(b"{") and line.endswith(b"}"), (
                        f"Line {line} does not start and end with a json object {{}}."
                    )
                    iw.append(last_offset)
                    last_offset = f.tell()
                    count += 1
                assert last_offset == f.tell(), (
                    f"The last offset {last_offset} does not match the file size {f.tell()}."
                )
                assert last_offset != 0, "File is empty."
                iw.append(last_offset)
        return count



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import field
from typing import Dict, List, Optional, Tuple, TypedDict

from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib import EPath
from megatron.energon.source_info import SourceInfo


@edataclass
class WebdatasetInfo:
    """Info about a webdataset. Format for `.nv-meta/.info.yaml` or `.nv-meta/.info.json`."""

    #: The version of the energon library that was used to prepare the dataset
    energon_version: Optional[str] = None
    #: Maps shard name to number of samples in that shard
    shard_counts: Dict[str, int]


@edataclass
class WebdatasetSplits:
    """Info about the splits of a webdataset. Format for `.nv-meta/split.yaml` or `.nv-meta/split.json`
    (or custom user yaml/json)."""

    #: Maps split part to list of shard names
    split_parts: Dict[str, List[str]]
    #: Set of "<shard name>" or "<shard name>/<sample index>" to exclude
    exclude: List[str] = field(default_factory=list)


@edataclass
class ShardInfo:
    """Info about a single shard as passed through internally. Not exposed to the user."""

    #: Name of the shard file (relative path from the nvinfo dir)
    name: str
    #: The path to the shard file
    path: EPath
    #: The number of samples in this shard
    count: int


class FilteredSample(TypedDict):
    """This is just a definition for the internal loaders. Not exposed to the user."""

    #: The key of the sample within the tar file.
    #: If the tar file contains files 12.jpg and 12.txt,
    #: those two files make one sample with the key "12"
    __key__: str
    #: The base name of the shard file e.g. "shard_000"
    __shard__: str
    #: Globally unique key to restore a sample from disk.
    #: For example `("Webdataset", 123)` would restore the sample at index 123.
    __restore_key__: Tuple[str, int]
    #: The source information for the sample.
    __sources__: tuple[SourceInfo, ...]


@edataclass
class DatasetSubset:
    """A subset of a dataset.
    A range is a tuple of two values, where the first value is the start of the subset and the second value is the end of the subset.

    The sharder uses the (absolute/relative) ranges to compute the subsets:
     * `absolute_range` (unit is samples) is applied first on the (e.g. train/val/test) subset
     * then `range` (where `(0, 1)` would correspond to the whole dataset) is applied as relative ratio on the subset that is left.

    This is the struct used internally for computing the range. The config is loaded via the metadataset_v2.
    """

    range: tuple[float, float] | None = None
    absolute_range: tuple[int, int | None] | None = None

    def compute_subset(
        self,
        total_samples: int,
    ) -> tuple[int, int]:
        """
        Computes the absolute subset of samples from the total number of samples.
        The absolute range is applied first, then the relative range is applied on the subset that is left.
        """
        start_samples = 0
        end_samples = total_samples

        if self.absolute_range is not None:
            start_samples, end_samples = self.absolute_range
            if end_samples is None:
                end_samples = total_samples
            assert end_samples <= total_samples, (
                f"Subset samples {self.absolute_range} {end_samples=} > {total_samples=}"
            )
            assert start_samples <= end_samples, (
                f"Subset samples {self.absolute_range} {start_samples=} > {end_samples=}"
            )
            assert start_samples >= 0, f"Subset samples {self.absolute_range} {start_samples=} < 0"
        if self.range is not None:
            previous_total = end_samples - start_samples
            end_samples = start_samples + int(previous_total * self.range[1])
            start_samples += int(previous_total * self.range[0])
            assert end_samples <= total_samples, (
                f"Subset ratio {self.range} {end_samples=} is larger than total samples {total_samples}"
            )
            assert start_samples <= end_samples, (
                f"Subset ratio {self.range} {start_samples=} > {end_samples=}"
            )
            assert start_samples >= 0, f"Subset ratio {self.range} {start_samples=} < 0"
        return start_samples, end_samples

    def config(self) -> dict:
        return {
            "range": self.range,
            "absolute_range": self.absolute_range,
        }


def reraise_exception(
    exc: Exception, key: Optional[str], sources: Optional[list[SourceInfo]] = None
) -> None:
    if sources:
        raise Exception(
            f"For sample {key!r} from {', '.join(f'{source.dataset_path}[{source.index}] {source.shard_name}{source.file_names!r}' for source in sources)}"
        ) from exc
    elif key:
        raise Exception(f"For sample {key!r}") from exc
    else:
        raise



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import json
import logging
from typing import Any, Callable, Dict, Optional

from megatron.energon.cache import FileStore
from megatron.energon.epathlib import EPath
from megatron.energon.flavors.base_dataset import (
    BaseCoreDatasetFactory,
    SavableDataset,
    legacy_handler,
)
from megatron.energon.flavors.crude import CrudeSample
from megatron.energon.flavors.jsonl.ijsonl import IJsonlIndexReader
from megatron.energon.flavors.jsonl.jsonl_prepare import JsonlPreparator
from megatron.energon.flavors.webdataset.base_webdataset import _print_shard_slices
from megatron.energon.flavors.webdataset.error_handler import ErrorHandler
from megatron.energon.flavors.webdataset.sample_loader import (
    RawSampleData,
    WebdatasetSampleLoaderDataset,
)
from megatron.energon.flavors.webdataset.sharder import Sharder
from megatron.energon.flavors.webdataset.structs import (
    DatasetSubset,
    FilteredSample,
    ShardInfo,
    reraise_exception,
)
from megatron.energon.source_info import SourceInfo
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.map_dataset import MapDataset

logger = logging.getLogger(__name__)


class CrudeJsonlDatasetFactory(
    BaseCoreDatasetFactory[CrudeSample],
    JsonlPreparator,
    Sharder,
    ErrorHandler,
):
    """
    Factory class for creating a crude dataset from JSONL (JSON Lines) files.

    This factory creates datasets from JSONL files where each line contains a JSON object.
    The samples are returned as CrudeSample objects (dictionary-like) containing the raw JSON data.
    """

    __sample_type__ = CrudeSample

    path: EPath

    training: bool
    worker_config: WorkerConfig

    def __init__(
        self,
        path: EPath,
        *,
        training: bool,
        worker_config: WorkerConfig,
        shuffle_over_epochs: Optional[int] = 1,
        parallel_shard_iters: Optional[int] = None,
        max_samples_per_sequence: Optional[int] = None,
        subset: Optional[DatasetSubset] = None,
        part_filter: Optional[Callable[[str], bool]] = None,
        handler: Callable[
            [Exception, Optional[str], Optional[list[SourceInfo]]], None
        ] = reraise_exception,
    ):
        """
        Factory for a jsonl file as a crude dataset.

        Args:
            path: Path to the jsonl file.
            training: If true, apply shuffling and loop the dataset.
            worker_config: Configuration for the workers.
            shuffle_over_epochs: Only effective if training=True.
                How many epochs to shuffle over if training.
                If = 1, every sample is seen exactly once per epoch.
                If > 1, samples (or rather shard slices) are shuffled within this number of epochs
                (i.e. randomly selected without replacement).
                If -1, the shards are effectively shuffle over infinite epochs (i.e. shard slices
                are drawn with replacement).
            parallel_shard_iters: Number of parallel opened shards per worker, shuffling between.
            max_samples_per_sequence: Maximum number of samples per sequence (=how many samples
                    will be sequentially iterated).
            subset: If specified, the dataset will be subsetted.
            part_filter: (internal) Function for filtering tar files by dict keys
            handler: Exception handler. Args: (exception, key).
        """
        assert self.__sample_type__ is not None, f"Class {type(self)} must define __sample_type__"
        self.path = path
        self.paths = [path]
        self.training = training
        self.worker_config = worker_config
        self.shuffle_over_epochs = shuffle_over_epochs
        self.parallel_shard_iters = parallel_shard_iters
        self.max_samples_per_sequence = max_samples_per_sequence
        self.subset = subset
        self.part_filter = part_filter
        self.handler = legacy_handler(handler)
        if part_filter is None or part_filter("json"):
            self._len = IJsonlIndexReader.count_samples(path)
        else:
            self._len = 0
        assert self.path.size() == IJsonlIndexReader.size(path), (
            "The index of the jsonl file does not match the file. Regenerate the index."
        )

    def __len__(self) -> int:
        return self._len

    def build(self, worker_rotation_offset: int = 0) -> SavableDataset[CrudeSample]:
        from megatron.energon.flavors.jsonl.ijsonl_reader import IJsonlReader

        if self.parallel_shard_iters is None:
            if self.training:
                # 16 seems to be a good choice since we don't want too many file handles open
                parallel_shard_iters = 16
            else:
                parallel_shard_iters = 1
        else:
            parallel_shard_iters = self.parallel_shard_iters

        virtual_shards = [
            ShardInfo(
                name=self.path.name,
                path=self.path,
                count=self._len,
            )
        ]

        workers_sample_slice_offsets = self.shard_workers(
            virtual_shards,
            worker_config=self.worker_config,
            max_samples_per_sequence=self.max_samples_per_sequence,
            rotation_offset=worker_rotation_offset,
            subset=self.subset,
        )
        _print_shard_slices(self.worker_config, virtual_shards, workers_sample_slice_offsets)

        itar_reader = IJsonlReader(
            self.path,
            index_cache_size=parallel_shard_iters,
        )

        dataset = WebdatasetSampleLoaderDataset(
            join_readers=[itar_reader],
            workers_sample_slice_offsets=workers_sample_slice_offsets,
            worker_config=self.worker_config,
            shuffle_over_epochs=self.shuffle_over_epochs if self.training else None,
            parallel_slice_iters=parallel_shard_iters,
        )
        return MapDataset(
            dataset,
            self._load_sample_raw,
            error_handler=self.error_handler,
            stateless_map_fn=True,
            map_fn_config=self.config,
            worker_config=self.worker_config,
        )

    def as_file_store(self) -> "FileStore":
        from megatron.energon.cache.file_store import JsonlFileStore

        return JsonlFileStore(self.path)

    def _load_sample(self, sample: FilteredSample) -> CrudeSample:
        return CrudeSample(sample)

    def _load_sample_raw(self, raw_sample: RawSampleData) -> CrudeSample:
        # Just a wrapper for the inner tuple. Tuple should be of length 1.
        assert len(raw_sample.data) == 1 and raw_sample.data[0] is not None
        return self._load_sample(raw_sample.data[0])

    def config(self) -> Dict[str, Any]:
        return dict(
            type=type(self).__qualname__,
            training=self.training,
            _path=str(self.path),
            jsonl_filename=self.path.name,
            count=self._len,
            shuffle_over_epochs=self.shuffle_over_epochs,
            parallel_shard_iters=self.parallel_shard_iters,
            max_samples_per_sequence=self.max_samples_per_sequence,
            subset=self.subset.config() if self.subset is not None else None,
        )

    def __str__(self):
        return f"{type(self).__name__}(path={self.path})"


class DefaultCrudeJsonlDatasetFactory(CrudeJsonlDatasetFactory):
    """
    Adds subflavors to the sample and loads the json.
    """

    def __init__(self, path: EPath, *, subflavors: Optional[Dict[str, Any]] = None, **kwargs):
        if "decoder" in kwargs:
            del kwargs["decoder"]
        super().__init__(path, **kwargs)
        self.subflavors = subflavors

    def _load_sample(self, sample: FilteredSample) -> CrudeSample:
        sample["__subflavors__"] = self.subflavors

        # Instead of using a decoder, we just load the json here, as we know it's json.
        sample["json"] = json.loads(sample["json"])

        return super()._load_sample(sample)

    def config(self) -> Dict[str, Any]:
        return dict(
            **super().config(),
            subflavors=self.subflavors,
        )



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import functools
import os
import weakref
from dataclasses import dataclass
from typing import Callable


def _cleanup(hooks, key, wr):
    hooks.pop(key)


class WeakCallbacks:
    """
    A class that manages weak references to callback functions.
    """

    # A dictionary of weak (or strong) references to functions.
    _hooks: dict[int, Callable[[], Callable[..., None] | None]]

    def __init__(self):
        """
        Initialize the registry.
        """
        self._hooks: dict[int, Callable[[], Callable[..., None] | None]] = {}

    def add_hook(self, callable: Callable[..., None], make_persistent: bool = False) -> None:
        """
        Add a callback to the registry.

        Args:
            callable: The function to run before the fork of a worker process.
            make_persistent: If True, the function will be stored as a strong reference, otherwise a weak reference is used.
        """
        if make_persistent:
            # Not a weakref, but always return the callable.
            self._hooks[id(callable)] = lambda: callable
        elif getattr(callable, "__self__", None):
            # Add a method reference to the hooks
            key = id(callable.__self__)
            self._hooks[key] = weakref.WeakMethod(
                callable, functools.partial(_cleanup, self._hooks, key)
            )
        else:
            # Add a function reference to the hooks
            key = id(callable)
            self._hooks[key] = weakref.ref(callable, functools.partial(_cleanup, self._hooks, key))

    def run(self, *args, **kwargs) -> None:
        """
        Run all the callbacks in the registry, passing the given arguments.
        """
        for hook in self._hooks.values():
            ref = hook()
            if ref is not None:
                ref(*args, **kwargs)


_after_in_child_fork_hooks = WeakCallbacks()
_after_in_parent_fork_hooks = WeakCallbacks()
_before_fork_hooks = WeakCallbacks()


def before_fork_hook(callable: Callable[[], None], make_persistent: bool = False):
    """
    Run function before the fork of a worker process.
    The function must be persistent (i.e. not a lambda) or an instance method.

    Args:
        callable: The function to run before the fork of a worker process.
        make_persistent: If True, the function will be stored as a strong reference, otherwise a weak reference is used.
    """
    _before_fork_hooks.add_hook(callable, make_persistent)


def after_in_parent_fork_hook(callable: Callable[[], None], make_persistent: bool = False):
    """
    Run function after the fork of a worker process.
    The function must be persistent (i.e. not a lambda) or an instance method.

    Args:
        callable: The function to run after the fork of a worker process.
        make_persistent: If True, the function will be stored as a strong reference, otherwise a weak reference is used.
    """
    _after_in_parent_fork_hooks.add_hook(callable, make_persistent)


def after_in_child_fork_hook(callable: Callable[[], None], make_persistent: bool = False):
    """
    Run function after the fork of a worker process.
    The function must be persistent (i.e. not a lambda) or an instance method.

    Args:
        callable: The function to run after the fork of a worker process.
        make_persistent: If True, the function will be stored as a strong reference, otherwise a weak reference is used.
    """
    _after_in_child_fork_hooks.add_hook(callable, make_persistent)


class ForkMixin:
    """
    A mixin that runs a method after the fork of a worker process.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__post_init__()

    def __post_init__(self):
        if getattr(self.__before_fork__, "__func__", None) is not ForkMixin.__before_fork__:
            before_fork_hook(self.__before_fork__)
        if (
            getattr(self.__after_in_child_fork__, "__func__", None)
            is not ForkMixin.__after_in_child_fork__
        ):
            after_in_child_fork_hook(self.__after_in_child_fork__)
        if (
            getattr(self.__after_in_parent_fork__, "__func__", None)
            is not ForkMixin.__after_in_parent_fork__
        ):
            after_in_parent_fork_hook(self.__after_in_parent_fork__)

    def __after_in_child_fork__(self):
        """
        A method that runs after the fork in the child process.
        """
        pass

    def __after_in_parent_fork__(self):
        """
        A method that runs after the fork in the parent process.
        """
        pass

    def __before_fork__(self):
        """
        A method that runs before the fork of a worker process.
        """
        pass


@dataclass
class DataclassForkMixin:
    """
    A mixin that runs a method after the fork of a worker process.
    """

    def __post_init__(self):
        if (
            getattr(self.__before_fork__, "__func__", None)
            is not DataclassForkMixin.__before_fork__
        ):
            before_fork_hook(self.__before_fork__)
        if (
            getattr(self.__after_in_child_fork__, "__func__", None)
            is not DataclassForkMixin.__after_in_child_fork__
        ):
            after_in_child_fork_hook(self.__after_in_child_fork__)
        if (
            getattr(self.__after_in_parent_fork__, "__func__", None)
            is not DataclassForkMixin.__after_in_parent_fork__
        ):
            after_in_parent_fork_hook(self.__after_in_parent_fork__)

    def __after_in_child_fork__(self):
        """
        A method that runs after the fork in the child process.
        """
        pass

    def __after_in_parent_fork__(self):
        """
        A method that runs after the fork in the parent process.
        """
        pass

    def __before_fork__(self):
        """
        A method that runs before the fork of a worker process.
        """
        pass


os.register_at_fork(
    before=_before_fork_hooks.run,
    after_in_child=_after_in_child_fork_hooks.run,
    after_in_parent=_after_in_parent_fork_hooks.run,
)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause
from megatron.energon.flavors.jsonl.crude_jsonl_dataset import (
    CrudeJsonlDatasetFactory,
    DefaultCrudeJsonlDatasetFactory,
)

__all__ = [
    "CrudeJsonlDatasetFactory",
    "DefaultCrudeJsonlDatasetFactory",
]
